import «02_lib»

namespace m114
open nat
/-
Dans la suite, on pourra utiliser les énoncés suivants.

inferieur_ssi (x y : ℝ) : x ≤ y ↔ 0 ≤ y - x

inferieur_ssi' (x y : ℝ) : x ≤ y ↔ x - y ≤ 0

pos_pos (x y : ℝ) (hx : 0 ≤ x) (hy : 0 ≤ y) : 0 ≤ x*y

neg_neg (x y : ℝ) (hx : x ≤ 0) (hy : y ≤ 0) : 0 ≤ x*y

Pour utiliser une équivalence, on peut réécrire comme avec une égalité,
à l'aide de rw et rw ←.

Pour démontrer une équivalence par double implication on utilise
split,

Pour démontrer une implication P → Q, la commande
intro hyp,
permet de supposer la prémisse P et d'appeler cette hypothèse hyp.

Pour introduire un énoncé intermédiaire, on utilise
have nom : énoncé,
Cela entraîne l'apparition d'un nouveau but à démontrer avant de reprendre le
fil de la démonstration principale.
-/

-- Dans toute la suite, a, b, c et d désignent des nombres réels.
variables {a b c d : ℝ}

lemma inferieur_add_gauche : a ≤ b ↔ c + a ≤ c + b :=
begin
  have clef : c + b - (c + a) = b - a,
    compute,
  split,
    intro hab,
    rw inferieur_ssi,
    rw clef,
    rw ← inferieur_ssi,
    exact hab,
  intro h,
  rw inferieur_ssi,
  rw ← clef,
  rw inferieur_ssi at h,
  exact h,
end

lemma inferieur_add_droite : a ≤ b ↔ a + c ≤ b + c :=
begin
  sorry
end

lemma inferieur_add_pos_gauche (hb : 0 ≤ a) : b ≤ a + b :=
begin
  calc b = 0 + b : by compute
     ... ≤ a + b : by { rw ← inferieur_add_droite, exact hb }
end

lemma inferieur_add_pos_droite (hb : 0 ≤ b) : a ≤ a + b :=
begin
  sorry
end

lemma inferieur_add_pos (ha : 0 ≤ a) (hb : 0 ≤ b) : 0 ≤ a + b :=
begin
  sorry
end

lemma inferieur_add (hab : a ≤ b) (hcd : c ≤ d) : a + c ≤ b + d :=
begin
  sorry
end

/-
Pour utiliser une implication on peut raisonner vers l'arrière : si P implique Q alors
pour montrer Q, il suffit de montrer P. La commande Lean pour faire cela avec une implication
imp est :
apply imp,
-/

lemma inferieur_mul_droite (hc : 0 ≤ c) (hab : a ≤ b) : a*c ≤ b*c :=
begin
  rw inferieur_ssi,
  have clef : b*c - a*c = (b - a)*c,
    by compute,
  rw clef,
  apply pos_pos, -- Noter que Lean comprend tout seul à quels nombres appliquer le lemme pos_pos
  { rw ← inferieur_ssi,
    exact hab },
  { exact hc },
end

-- Le même en raisonnant vers l'avant
example (hc : 0 ≤ c) (hab : a ≤ b) : a*c ≤ b*c :=
begin
  have hab' : 0 ≤ b - a,
    rw ← inferieur_ssi,
    exact hab,
  have h₁ : 0 ≤ (b - a)*c,
    exact pos_pos hab' hc,
  have h₂ : (b - a)*c = b*c - a*c,
    compute,
  have h₃ : 0 ≤ b*c - a*c,
    rw ← h₂,
    exact h₁,
  rw inferieur_ssi,
  exact h₃,
end

-- Encore une variante avec un calc
example (hc : 0 ≤ c) (hab : a ≤ b) : a*c ≤ b*c :=
begin
  have hab' : 0 ≤ b - a,
    rw ← inferieur_ssi,
    exact hab,
  rw inferieur_ssi,
  calc 0 ≤ (b - a)*c  : by exact pos_pos hab' hc
     ... =  b*c - a*c : by compute,
end

example (hc : c ≤ 0) (hab :  a ≤ b) : b*c ≤ a*c :=
begin
  sorry
end

-- Essayer maintenant un autre style pour démontrer le même énoncé
example (hc : c ≤ 0) (hab :  a ≤ b) : b*c ≤ a*c :=
begin
  sorry
end

/-
Beaucoup d'énoncés ci-dessus cachent en fait des implications.
Par exemple, l'énoncé
lemma inferieur_add_pos_gauche (hb : 0 ≤ a) : b ≤ a + b
se lit « Supposons que 0 ≤ a. Montrons que b ≤ a + b ». Mais on peut le réécrire
lemma inferieur_add_pos_gauche' : 0 ≤ a → b ≤ a + b
qui se lit « Si 0 ≤ a alors b ≤ a + b ». Le contenu mathématique de ces deux énoncés est le même,
la différence est purement stylistique.

Vérifions qu'on peut ramener se ramener à la première version.
-/
lemma inferieur_add_pos_gauche' : 0 ≤ a → b ≤ a + b :=
begin
  intro ha,
  exact inferieur_add_pos_gauche ha,
end

/-
En fait Lean ne fait aucune différence entre les deux versions de ce lemme, comme on peut le voir
dans la version suivante.
-/
lemma inferieur_add_pos_gauche'' : 0 ≤ a → b ≤ a + b :=
begin
  exact inferieur_add_pos_gauche,
end

/-
On se demande alors naturellement comment reformuler les énoncés faisant *deux* hypothèses, comme
lemma inferieur_add_pos (ha : 0 ≤ a) (hb : 0 ≤ b) : 0 ≤ a + b

La solution la plus intuitive est de faire appel au connecteur logique « et », noté « ∧ ».
Les hypothèses du contexte local qui sont constituées de plusieurs morceaux peuvent être scindées
à l'aide de la commande « cases » comme dans l'exemple suivant.
-/
lemma inferieur_add_pos' : (0 ≤ a ∧ 0 ≤ b) → 0 ≤ a + b :=
begin
  intro hyp,
  cases hyp with ha hb,
  exact inferieur_add_pos ha hb,
end

/-
La nécessité de passer par une ligne intermédiaire qui scinde la conjonction hyp en ha et hb
montre que cette reformulation comme implication n'est pas celle que Lean utilise.
Lean reformule l'énoncé comme deux implications emboîtées, ce qui est plus commode à l'usage.
-/
lemma inferieur_add_pos'' : 0 ≤ a → (0 ≤ b → 0 ≤ a + b) :=
begin
  exact inferieur_add_pos,
end

/-
Montrons que la version utilisant une conjonction permet de retrouver la version de Lean.
Pour cela il faut savoir démontrer un énoncé de la forme « P et Q » à l'aide de la commande
split,
qui scinde la démonstration en deux buts P et Q.
-/
lemma inferieur_add_pos''' : 0 ≤ a → (0 ≤ b → 0 ≤ a + b) :=
begin
  intro ha,
  intro hb,
  apply inferieur_add_pos',
  split,
  exact ha,
  exact hb,
end

/-
L'équivalence entre les deux reformulations est un fait complètement général.
Avant de démontrer ce fait, voici un exemple plus simple pour manipuler
cases et split.
Dans cet exercice, P, Q et R désignent des énoncés mathématiques quelconques.
-/

example (P Q R : Prop): P ∧ Q → Q ∧ P :=
begin
  sorry
end

/-
Revenons maintenant aux deux façons de formuler des énoncés faisant deux hypothèses.
-/
example (P Q R : Prop) : (P ∧ Q → R) ↔ (P → (Q → R)) :=
begin
  sorry
end

/-
Final

On a la relation de divisibilité sur les entiers naturels :
a ∣ b si b est un multiple de a. Attention la barre verticale n'est pas celle du
clavier mais s'obtient par ,|

Cette relation est réfléxive et antisymétrique, c'est à dire qu'on a les deux énoncés :
divise_refl (a : ℕ) : a ∣ a
et
divise_antisym {a b : ℕ} : a ∣ b → b ∣ a → a = b

De plus on a une fonction pgcd qui, à deux entiers naturels associde un troisième entier
(leur plus grand diviseur commun). Cette fonction est liée à la relation
de divisibilité par :

divise_pgcd_ssi {a b c : ℕ} : c ∣ pgcd a b ↔ c ∣ a ∧ c ∣ b
-/

example (a b : ℕ) : a ∣ b ↔ pgcd a b = a :=
begin
  sorry
end
end m114

